@@ -67,10 +67,19 @@
 #define VELOCITY_LOW		32
 #define VELOCITY_MIN		1
 
+#define MOD_MIN				0
+#define MOD_MAX				127
+#define MOD_INIT			0
+
+#define PITCH_MIN			-100
+#define PITCH_MAX			100
+#define PITCH_INIT			0
+#define PITCH_RANGE			8192
+
 #define OUTPUT_PORT_NAME	"midi_out"
 #define INPUT_PORT_NAME		"midi_in"
 #define PACKAGE_NAME		"jack-keyboard"
-#define PACKAGE_VERSION		"2.7.1"
+#define PACKAGE_VERSION		"2.7.2"
 
 jack_port_t	*output_port;
 jack_port_t	*input_port;
@@ -96,10 +105,13 @@ jack_client_t	*jack_client = NULL;
 lash_client_t	*lash_client;
 #endif
 
+static void panic(void);
+
 #define MIDI_NOTE_ON		0x90
 #define MIDI_NOTE_OFF		0x80
 #define MIDI_PROGRAM_CHANGE	0xC0
 #define MIDI_CONTROLLER		0xB0
+#define MIDI_PITCH		0xE0
 #define MIDI_RESET		0xFF
 #define MIDI_HOLD_PEDAL		64
 #define MIDI_ALL_SOUND_OFF	120
@@ -107,6 +119,7 @@ lash_client_t	*lash_client;
 #define MIDI_ALL_NOTES_OFF	123
 #define MIDI_BANK_SELECT_MSB	0
 #define MIDI_BANK_SELECT_LSB	32
+#define MIDI_MOD_CC		1
 
 #define BANK_MIN		0
 #define BANK_MAX		127
@@ -115,8 +128,58 @@ lash_client_t	*lash_client;
 #define CHANNEL_MIN		1
 #define CHANNEL_MAX		16
 
+const char *rc_style = "style \"default\"\n"
+"{\n"
+"	xthickness		 = 3\n"
+"	ythickness		 = 2\n"
+"	fg[NORMAL]		= \"#fff\"\n"
+"	fg[PRELIGHT]	  = \"#fff\"\n"
+"	fg[ACTIVE]		= \"#fff\"\n"
+"	fg[SELECTED]	  = \"#fff\"\n"
+"	fg[INSENSITIVE]   = \"#aaa\"\n"
+"	bg[NORMAL]		= \"#222\"\n"
+"	bg[PRELIGHT]	  = \"#444\"\n"
+"	bg[ACTIVE]		= \"#222\"\n"
+"	bg[SELECTED]	  = \"#0af\"\n"
+"	bg[INSENSITIVE]   = \"#666\"\n"
+"	base[NORMAL]	  = \"#000\"\n"
+"	base[PRELIGHT]	= \"#444\"\n"
+"	base[ACTIVE]	  = \"#444\"\n"
+"	base[SELECTED]	= \"#444\"\n"
+"	base[INSENSITIVE] = \"#333\"\n"
+"	text[NORMAL]	  = \"#fff\"\n"
+"	text[PRELIGHT]	= \"#fff\"\n"
+"	text[ACTIVE]	  = \"#f80\"\n"
+"	text[SELECTED]	= \"#f80\"\n"
+"	text[INSENSITIVE] = \"#666\"\n"
+"}\n"
+"style \"button\"\n"
+"{\n"
+"	fg[NORMAL]		= \"#fff\"\n"
+"	bg[NORMAL]		= \"#000\"\n"
+"	bg[PRELIGHT]	  = \"#333\"\n"
+"	bg[ACTIVE]		= \"#444\"\n"
+"	bg[INSENSITIVE]   = \"#222\"\n"
+"}\n"
+"style \"scale\"\n"
+"{\n"
+"	bg[NORMAL]		= \"#555\"\n"
+"	bg[PRELIGHT]	  = \"#000\"\n"
+"	bg[ACTIVE]		= \"#000\"\n"
+"	bg[INSENSITIVE]   = \"#333\"\n"
+"}\n"
+"style \"label\"\n"
+"{\n"
+"	fg[NORMAL]		= \"#aaa\"\n"
+"}\n"
+"widget \"*\" style : rc \"default\"\n"
+"widget_class\"*Gtk*Scale*\" style : highest \"scale\"\n"
+"widget_class\"*GtkLabel*\" style : highest \"label\"\n"
+"widget_class\"*GtkButton*\" style : highest \"button\"\n"
+"widget_class\"*GtkToggle*\" style : highest \"button\"\n";
+
 GtkWidget	*window, *sustain_button, *channel_spin, *bank_spin, *program_spin, *connected_to_combo,
-		*velocity_hscale, *grab_keyboard_checkbutton, *octave_spin;
+		*velocity_scale, *grab_keyboard_checkbutton, *octave_spin, *mod_scale, *pitch_scale, *panic_button;
 PianoKeyboard	*keyboard;
 GtkListStore	*connected_to_store;
 
@@ -152,7 +215,7 @@ int		channel = 0;
 void draw_note(int key);
 void queue_message(struct MidiMessage *ev);
 
-double 
+double
 get_time(void)
 {
 	double seconds;
@@ -218,7 +281,7 @@ process_received_message_async(gpointer evp)
 		if (ev->data[2] == 0)
 			piano_keyboard_set_note_off(keyboard, ev->data[1]);
 		else
-			piano_keyboard_set_note_on(keyboard, ev->data[1]);
+			piano_keyboard_set_note_on(keyboard, ev->data[1], ev->data[2]);
 	}
 
 	if (b0 == MIDI_NOTE_OFF) {
@@ -256,7 +319,7 @@ warning_async(gpointer s)
 {
 	const char *str = (const char *)s;
 
-	g_warning(str);
+	g_warning("%s", str);
 
 	return (FALSE);
 }
@@ -399,7 +462,7 @@ process_midi_output(jack_nframes_t nframes)
 	}
 }
 
-int 
+int
 process_callback(jack_nframes_t nframes, void *notused)
 {
 #ifdef MEASURE_TIME
@@ -440,7 +503,7 @@ queue_message(struct MidiMessage *ev)
 		g_warning("jack_ringbuffer_write failed, NOTE LOST.");
 }
 
-void 
+void
 queue_new_message(int b0, int b1, int b2)
 {
 	struct MidiMessage ev;
@@ -524,7 +587,7 @@ update_connected_to_combo_async(gpointer notused)
 	return (FALSE);
 }
 
-void 
+void
 draw_window_title(void)
 {
 	int i, off = 0;
@@ -577,7 +640,7 @@ update_window_title_async(gpointer notused)
 	return (FALSE);
 }
 
-int 
+int
 graph_order_callback(void *notused)
 {
 	g_idle_add(update_window_title_async, NULL);
@@ -625,7 +688,7 @@ connect_to_input_port(const char *port)
                 g_warning("Cannot connect to %s.", port);
 
                 return (-4);
-        } 
+        }
 
 	g_warning("Connected to %s.", port);
 
@@ -725,19 +788,19 @@ connected:
 		free(connected_ports);
 }
 
-void 
+void
 connect_to_next_input_port(void)
 {
 	connect_to_another_input_port(1);
 }
 
-void 
+void
 connect_to_prev_input_port(void)
 {
 	connect_to_another_input_port(0);
 }
 
-void 
+void
 init_jack(void)
 {
 	int err;
@@ -815,7 +878,7 @@ load_config_from_lash(void)
 	int value;
 
 	while ((config = lash_get_config(lash_client))) {
-		
+
 		key = lash_config_get_key(config);
 		value = lash_config_get_value_int(config);
 
@@ -859,7 +922,7 @@ load_config_from_lash(void)
 				g_warning("Bad value '%d' for 'velocity_normal' property received from LASH.", value);
 			} else {
 		                velocity_normal = value;
-				gtk_range_set_value(GTK_RANGE(velocity_hscale), *current_velocity);
+				gtk_range_set_value(GTK_RANGE(velocity_scale), *current_velocity);
 			}
 
 		} else if (!strcmp(key, "velocity_high")) {
@@ -867,7 +930,7 @@ load_config_from_lash(void)
 				g_warning("Bad value '%d' for 'velocity_high' property received from LASH.", value);
 			} else {
 		                velocity_high = value;
-				gtk_range_set_value(GTK_RANGE(velocity_hscale), *current_velocity);
+				gtk_range_set_value(GTK_RANGE(velocity_scale), *current_velocity);
 			}
 
 		} else if (!strcmp(key, "velocity_low")) {
@@ -875,7 +938,7 @@ load_config_from_lash(void)
 				g_warning("Bad value '%d' for 'velocity_low' property received from LASH.", value);
 			} else {
 		                velocity_low = value;
-				gtk_range_set_value(GTK_RANGE(velocity_hscale), *current_velocity);
+				gtk_range_set_value(GTK_RANGE(velocity_scale), *current_velocity);
 			}
 
 		} else {
@@ -1028,11 +1091,32 @@ velocity_event_handler(GtkRange *range, gpointer notused)
 	assert(current_velocity);
 
 	*current_velocity = gtk_range_get_value(range);
+	keyboard->current_velocity = gtk_range_get_value(range);
+}
+
+void
+mod_event_handler(GtkRange *range, gpointer notused)
+{
+	int val = (int) gtk_range_get_value(range);
+	queue_new_message(MIDI_CONTROLLER, MIDI_MOD_CC, val);
+}
+
+void
+pitch_event_handler(GtkRange *range, gpointer notused)
+{
+	uint16_t val = (uint16_t) (gtk_range_get_value(range) * ((float)PITCH_RANGE / (float)PITCH_MAX) + (float)PITCH_RANGE);
+	queue_new_message(MIDI_PITCH, val & 127, (val >> 7) & 127);
+}
+
+void
+panic_event_handler(GtkWidget *widget)
+{
+	panic();
 }
 
 #ifdef HAVE_X11
 
-int grab_x_error_handler(Display *dpy, XErrorEvent *notused) 
+int grab_x_error_handler(Display *dpy, XErrorEvent *notused)
 {
 	keyboard_grabbed = 0;
 
@@ -1085,7 +1169,7 @@ grab_keyboard(void)
 		XK_q, XK_w, XK_e, XK_r, XK_t, XK_y, XK_u, XK_i, XK_o, XK_p,
 		XK_a, XK_s, XK_d, XK_f, XK_g, XK_h, XK_j, XK_k, XK_l,
 		XK_z, XK_x, XK_c, XK_v, XK_b, XK_n, XK_m,
-		XK_Shift_L, XK_Shift_R, XK_Control_L, XK_Control_R, XK_space, 
+		XK_Shift_L, XK_Shift_R, XK_Control_L, XK_Control_R, XK_space,
 		XK_Insert, XK_Delete, XK_Home, XK_End, XK_Page_Up, XK_Page_Down,
 		XK_KP_1, XK_KP_2, XK_KP_3, XK_KP_4, XK_KP_5, XK_KP_6, XK_KP_7, XK_KP_8, XK_KP_9, XK_KP_0,
 		XK_Num_Lock, XK_KP_Multiply, XK_KP_Divide, XK_KP_Subtract, XK_KP_Add, XK_KP_Enter,
@@ -1169,7 +1253,7 @@ octave_event_handler(GtkSpinButton *spinbutton, gpointer notused)
 	piano_keyboard_set_octave(keyboard, octave);
 }
 
-void 
+static void
 panic(void)
 {
 	int i;
@@ -1314,7 +1398,7 @@ clip(int val, int lo, int hi)
 	return (val);
 }
 
-gint 
+gint
 keyboard_event_handler(GtkWidget *widget, GdkEventKey *event, gpointer notused)
 {
 	int tmp;
@@ -1493,7 +1577,7 @@ keyboard_event_handler(GtkWidget *widget, GdkEventKey *event, gpointer notused)
 		else
 			current_velocity = &velocity_normal;
 
-		gtk_range_set_value(GTK_RANGE(velocity_hscale), *current_velocity);
+		gtk_range_set_value(GTK_RANGE(velocity_scale), *current_velocity);
 
 		return (TRUE);
 
@@ -1505,7 +1589,7 @@ keyboard_event_handler(GtkWidget *widget, GdkEventKey *event, gpointer notused)
 		else
 			current_velocity = &velocity_normal;
 
-		gtk_range_set_value(GTK_RANGE(velocity_hscale), *current_velocity);
+		gtk_range_set_value(GTK_RANGE(velocity_scale), *current_velocity);
 
 		return (TRUE);
 	}
@@ -1550,54 +1634,78 @@ init_gtk_1(int *argc, char ***argv)
 	/* Window. */
 	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
 	gtk_container_set_border_width(GTK_CONTAINER(window), 2);
+	
+	gtk_rc_parse_string (rc_style);
 }
 
 void
 init_gtk_2(void)
 {
-	GtkTable *table;
+	keyboard = PIANO_KEYBOARD(piano_keyboard_new());
+	
+	if (!enable_gui) {
+		gtk_container_add(GTK_CONTAINER(window), GTK_WIDGET(keyboard));
+		return;
+	}
+	
 	GtkWidget *label;
-	GtkCellRenderer *renderer;
-
-	/* Table. */
-	table = GTK_TABLE(gtk_table_new(4, 8, FALSE));
-	gtk_table_set_row_spacings(table, 5);
-	gtk_table_set_col_spacings(table, 5);
-
-	if (enable_gui)
-		gtk_container_add(GTK_CONTAINER(window), GTK_WIDGET(table));
+	
+	/* Main Table. */
+	GtkTable *maintable = GTK_TABLE(gtk_table_new(3, 4, FALSE));
+	gtk_table_set_row_spacings(maintable, 5);
+	gtk_table_set_col_spacings(maintable, 5);
+	gtk_container_add(GTK_CONTAINER(window), GTK_WIDGET(maintable));
+	
+	
+	/* Num Table. */
+	GtkWidget *numwidget = gtk_table_new(2, 4, FALSE);
+	gtk_table_set_row_spacings(GTK_TABLE(numwidget), 5);
+	gtk_table_set_col_spacings(GTK_TABLE(numwidget), 5);
+	gtk_table_attach(maintable, numwidget, 1, 2, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
 
 	/* Channel label and spin. */
-	label = gtk_label_new("Channel:");
-	gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
-	gtk_table_attach(table, label, 0, 1, 0, 1, GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
+	label = gtk_label_new("Chan");
+	gtk_table_attach(GTK_TABLE(numwidget), label, 0, 1, 0, 1,GTK_FILL, GTK_FILL, 0, 0);
 	channel_spin = gtk_spin_button_new_with_range(1, CHANNEL_MAX, 1);
 	GTK_WIDGET_UNSET_FLAGS(channel_spin, GTK_CAN_FOCUS);
-	gtk_table_attach(table, channel_spin, 1, 2, 0, 1, GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
+	gtk_table_attach(GTK_TABLE(numwidget), channel_spin, 0, 1, 1, 2, GTK_FILL, GTK_FILL, 0, 0);
 	g_signal_connect(G_OBJECT(channel_spin), "value-changed", G_CALLBACK(channel_event_handler), NULL);
-
+	
+	/* Octave label and spin. */
+	label = gtk_label_new("Oct");
+	gtk_table_attach(GTK_TABLE(numwidget), label, 1, 2, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
+	octave_spin = gtk_spin_button_new_with_range(OCTAVE_MIN, OCTAVE_MAX, 1);
+	GTK_WIDGET_UNSET_FLAGS(octave_spin, GTK_CAN_FOCUS);
+	gtk_spin_button_set_value(GTK_SPIN_BUTTON(octave_spin), octave);
+	gtk_table_attach(GTK_TABLE(numwidget), octave_spin, 1, 2, 1, 2, GTK_FILL, GTK_FILL, 0, 0);
+	g_signal_connect(G_OBJECT(octave_spin), "value-changed", G_CALLBACK(octave_event_handler), NULL);
+	
 	/* Bank label and spin. */
-	label = gtk_label_new("Bank:");
-	gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
-	gtk_table_attach(table, label, 2, 3, 0, 1, GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
+	label = gtk_label_new("Bank");
+	gtk_table_attach(GTK_TABLE(numwidget), label, 2, 3, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
 	bank_spin = gtk_spin_button_new_with_range(0, BANK_MAX, 1);
 	GTK_WIDGET_UNSET_FLAGS(bank_spin, GTK_CAN_FOCUS);
-	gtk_table_attach(table, bank_spin, 3, 4, 0, 1, GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
+	gtk_table_attach(GTK_TABLE(numwidget), bank_spin, 2, 3, 1, 2, GTK_FILL, GTK_FILL, 0, 0);
 	g_signal_connect(G_OBJECT(bank_spin), "value-changed", G_CALLBACK(bank_event_handler), NULL);
 
 	/* Program label and spin. */
-	label = gtk_label_new("Program:");
-	gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
-	gtk_table_attach(table, label, 4, 5, 0, 1, GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
+	label = gtk_label_new("Prog");
+	gtk_table_attach(GTK_TABLE(numwidget), label, 3, 4, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
 	program_spin = gtk_spin_button_new_with_range(0, PROGRAM_MAX, 1);
 	GTK_WIDGET_UNSET_FLAGS(program_spin, GTK_CAN_FOCUS);
-	gtk_table_attach(table, program_spin, 5, 6, 0, 1, GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
+	gtk_table_attach(GTK_TABLE(numwidget), program_spin, 3, 4, 1, 2, GTK_FILL, GTK_FILL, 0, 0);
 	g_signal_connect(G_OBJECT(program_spin), "value-changed", G_CALLBACK(program_event_handler), NULL);
-
+	
+	/* PANIC */
+	panic_button = gtk_button_new_with_label("    PANIC    ");
+	gtk_button_set_focus_on_click(GTK_BUTTON(panic_button), FALSE);
+	gtk_table_attach(maintable, panic_button, 0, 1, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
+	g_signal_connect(G_OBJECT(panic_button), "pressed", G_CALLBACK(panic_event_handler), NULL);
+	
 	/* "Connected to" label and combo box. */
-	label = gtk_label_new("Connected to:");
-	gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
-	gtk_table_attach(table, label, 6, 7, 0, 1, GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
+	GtkCellRenderer *renderer;
+	label = gtk_label_new("Connect");
+	gtk_table_attach(GTK_TABLE(numwidget), label, 4, 5, 0, 1, GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0);
 
 	connected_to_store = gtk_list_store_new(1, G_TYPE_STRING);
 	connected_to_combo = gtk_combo_box_new_with_model(GTK_TREE_MODEL(connected_to_store));
@@ -1608,57 +1716,78 @@ init_gtk_2(void)
 
 	GTK_WIDGET_UNSET_FLAGS(connected_to_combo, GTK_CAN_FOCUS);
 	gtk_combo_box_set_focus_on_click(GTK_COMBO_BOX(connected_to_combo), FALSE);
-	gtk_table_attach(table, connected_to_combo, 7, 8, 0, 1, GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
-	gtk_widget_set_size_request(GTK_WIDGET(connected_to_combo), 200, -1);
+	gtk_table_attach(GTK_TABLE(numwidget), connected_to_combo, 4, 5, 1, 2, GTK_FILL, GTK_FILL, 0, 0);
+	gtk_widget_set_size_request(GTK_WIDGET(connected_to_combo), 160, -1);
 	g_signal_connect(G_OBJECT(connected_to_combo), "changed", G_CALLBACK(connected_to_event_handler), NULL);
 
-	/* Octave label and spin. */
-	label = gtk_label_new("Octave:");
-	gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
-	gtk_table_attach(table, label, 0, 1, 1, 2, GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
-	octave_spin = gtk_spin_button_new_with_range(OCTAVE_MIN, OCTAVE_MAX, 1);
-	GTK_WIDGET_UNSET_FLAGS(octave_spin, GTK_CAN_FOCUS);
-	gtk_spin_button_set_value(GTK_SPIN_BUTTON(octave_spin), octave);
-	gtk_table_attach(table, octave_spin, 1, 2, 1, 2, GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
-	g_signal_connect(G_OBJECT(octave_spin), "value-changed", G_CALLBACK(octave_event_handler), NULL);
-
 	/* "Grab keyboard" label and checkbutton. */
-	label = gtk_label_new("Grab keyboard:");
-	gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
-	gtk_table_attach(table, label, 4, 5, 1, 2, GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
+	label = gtk_label_new("Grab Keyboard");
+	gtk_table_attach(GTK_TABLE(numwidget), label, 5, 6, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
 	grab_keyboard_checkbutton = gtk_check_button_new();
 	GTK_WIDGET_UNSET_FLAGS(grab_keyboard_checkbutton, GTK_CAN_FOCUS);
 	g_signal_connect(G_OBJECT(grab_keyboard_checkbutton), "toggled", G_CALLBACK(grab_keyboard_handler), NULL);
-	gtk_table_attach(table, grab_keyboard_checkbutton, 5, 6, 1, 2, GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
-
-	/* Velocity label and hscale */
-	label = gtk_label_new("Velocity:");
-	gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
-	gtk_table_attach(table, label, 6, 7, 1, 2, GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
-
-	velocity_hscale = gtk_hscale_new_with_range(VELOCITY_MIN, VELOCITY_MAX, 1);
-	gtk_scale_set_draw_value(GTK_SCALE(velocity_hscale), FALSE);
-	GTK_WIDGET_UNSET_FLAGS(velocity_hscale, GTK_CAN_FOCUS);
-	g_signal_connect(G_OBJECT(velocity_hscale), "value-changed", G_CALLBACK(velocity_event_handler), NULL);
-	gtk_range_set_value(GTK_RANGE(velocity_hscale), VELOCITY_NORMAL);
-	gtk_table_attach(table, velocity_hscale, 7, 8, 1, 2, GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
-	gtk_widget_set_size_request(GTK_WIDGET(velocity_hscale), 200, -1);
+	gtk_table_attach(GTK_TABLE(numwidget), grab_keyboard_checkbutton, 5, 6, 1, 2, GTK_FILL, GTK_FILL, 0, 0);
 
 	/* Sustain. It's a toggle button, not an ordinary one, because we want gtk_whatever_set_active() to work.*/
 	sustain_button = gtk_toggle_button_new_with_label("Sustain");
 	gtk_button_set_focus_on_click(GTK_BUTTON(sustain_button), FALSE);
-	gtk_table_attach(table, sustain_button, 0, 8, 2, 3, GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
+	gtk_table_attach(maintable, sustain_button, 0, 3, 1, 2, GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
 	g_signal_connect(G_OBJECT(sustain_button), "pressed", G_CALLBACK(sustain_event_handler), (void *)1);
 	g_signal_connect(G_OBJECT(sustain_button), "released", G_CALLBACK(sustain_event_handler), (void *)0);
 
-	/* PianoKeyboard widget. */
-	keyboard = PIANO_KEYBOARD(piano_keyboard_new());
+	/* Mod Table. */
+	GtkWidget *modwidget = gtk_table_new(2, 3, FALSE);
+	gtk_table_set_row_spacings(GTK_TABLE(modwidget), 2);
+	gtk_table_set_col_spacings(GTK_TABLE(modwidget), 5);
+	gtk_table_attach(maintable, modwidget, 3, 4, 0, 3, GTK_FILL, GTK_FILL, 0, 0);
+	
+	/* Velocity label and hscale */
+	label = gtk_label_new("Vel");
+	gtk_table_attach(GTK_TABLE(modwidget), label, 0, 1, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
 
-	if (enable_gui)
-		gtk_table_attach_defaults(table, GTK_WIDGET(keyboard), 0, 8, 3, 4);
-	else
-		gtk_container_add(GTK_CONTAINER(window), GTK_WIDGET(keyboard));
+	velocity_scale = gtk_vscale_new_with_range(VELOCITY_MIN, VELOCITY_MAX, 1);
+	gtk_scale_set_value_pos(GTK_SCALE(velocity_scale), GTK_POS_BOTTOM);
+	GTK_WIDGET_UNSET_FLAGS(velocity_scale, GTK_CAN_FOCUS);
+	g_signal_connect(G_OBJECT(velocity_scale), "value-changed", G_CALLBACK(velocity_event_handler), NULL);
+	gtk_range_set_value(GTK_RANGE(velocity_scale), VELOCITY_NORMAL);
+	gtk_range_set_inverted(GTK_RANGE(velocity_scale), TRUE);
+	gtk_range_set_round_digits(GTK_RANGE(velocity_scale), 0);
+	gtk_table_attach(GTK_TABLE(modwidget), velocity_scale, 0, 1, 1, 2, GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
+	gtk_widget_set_size_request(GTK_WIDGET(velocity_scale), 40, 160);
+	
+	/* Modulation label and vscale */
+	label = gtk_label_new("Mod");
+	gtk_table_attach(GTK_TABLE(modwidget), label, 1, 2, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
 
+	mod_scale = gtk_vscale_new_with_range(MOD_MIN, MOD_MAX, 1);
+	gtk_scale_set_value_pos(GTK_SCALE(mod_scale), GTK_POS_BOTTOM);
+	GTK_WIDGET_UNSET_FLAGS(mod_scale, GTK_CAN_FOCUS);
+	gtk_range_set_value(GTK_RANGE(mod_scale), MOD_INIT);
+	gtk_range_set_inverted(GTK_RANGE(mod_scale), TRUE);
+	gtk_range_set_round_digits(GTK_RANGE(mod_scale), 0);
+	g_signal_connect(G_OBJECT(mod_scale), "value-changed", G_CALLBACK(mod_event_handler), NULL);
+	gtk_table_attach(GTK_TABLE(modwidget), mod_scale, 1, 2, 1, 2, GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
+	gtk_widget_set_size_request(GTK_WIDGET(mod_scale), 40, 160);
+	
+	/* Pitch Bend label and vscale */
+	label = gtk_label_new("Pitch");
+	gtk_table_attach(GTK_TABLE(modwidget), label, 2, 3, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
+
+	pitch_scale = gtk_vscale_new_with_range(PITCH_MIN, PITCH_MAX, 1);
+	gtk_scale_set_value_pos(GTK_SCALE(pitch_scale), GTK_POS_BOTTOM);
+	GTK_WIDGET_UNSET_FLAGS(pitch_scale, GTK_CAN_FOCUS);
+	gtk_range_set_value(GTK_RANGE(pitch_scale), PITCH_INIT);
+	gtk_range_set_inverted(GTK_RANGE(pitch_scale), TRUE);
+	gtk_range_set_round_digits(GTK_RANGE(pitch_scale), 0);
+	g_signal_connect(G_OBJECT(pitch_scale), "value-changed", G_CALLBACK(pitch_event_handler), NULL);
+	gtk_table_attach(GTK_TABLE(modwidget), pitch_scale, 2, 3, 1, 2, GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
+	gtk_widget_set_size_request(GTK_WIDGET(pitch_scale), 40, 160);
+	
+	
+	/* Keyboard */
+	gtk_table_attach(maintable, GTK_WIDGET(keyboard), 0, 3, 2, 3, GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
+	
+	
 	g_signal_connect(G_OBJECT(keyboard), "note-on", G_CALLBACK(note_on_event_handler), NULL);
 	g_signal_connect(G_OBJECT(keyboard), "note-off", G_CALLBACK(note_off_event_handler), NULL);
 	g_signal_connect(G_OBJECT(window), "destroy", G_CALLBACK(gtk_main_quit), NULL);
@@ -1669,7 +1798,7 @@ init_gtk_2(void)
 	draw_window_title();
 }
 
-void   
+void
 log_handler(const gchar *log_domain, GLogLevelFlags log_level, const gchar *message, gpointer notused)
 {
 	GtkWidget *dialog;
@@ -1678,7 +1807,7 @@ log_handler(const gchar *log_domain, GLogLevelFlags log_level, const gchar *mess
 
 	if ((log_level | G_LOG_LEVEL_CRITICAL) == G_LOG_LEVEL_CRITICAL) {
 		dialog = gtk_message_dialog_new(GTK_WINDOW(window), GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_ERROR,
-			GTK_BUTTONS_CLOSE, message);
+			GTK_BUTTONS_CLOSE, "%s", message);
 
 		gtk_dialog_run(GTK_DIALOG(dialog));
 
@@ -1701,13 +1830,13 @@ usage(void)
 	fprintf(stderr, "   where <channel> is MIDI channel to use for output, from 1 to 16,\n");
 	fprintf(stderr, "   <bank> is MIDI bank to use, from 0 to 16383,\n");
 	fprintf(stderr, "   <program> is MIDI program to use, from 0 to 127,\n");
-	fprintf(stderr, "   and <layout> is  QWERTY, QWERTY_REV, QWERTY_UK, QWERTY_UK_REV, QWERTZ, AZERTY or DVORAK.\n");
+	fprintf(stderr, "   and <layout> is QWERTY.\n");
 	fprintf(stderr, "See manual page for details.\n");
 
 	exit(EX_USAGE);
 }
 
-int 
+int
 main(int argc, char *argv[])
 {
 	int ch, enable_keyboard_cue = 0, initial_channel = 1, initial_bank = 0, initial_program = 0, full_midi_keyboard = 0;
@@ -1765,11 +1894,11 @@ main(int argc, char *argv[])
 		case 'l':
 			keyboard_layout = strdup(optarg);
 			break;
-			
+
 		case 't':
 			time_offsets_are_zero = 1;
 			break;
-			
+
 		case 'u':
 			send_program_change_at_reconnect = 1;
 			break;
@@ -1846,7 +1975,7 @@ main(int argc, char *argv[])
 		int ret = piano_keyboard_set_keyboard_layout(keyboard, keyboard_layout);
 
 		if (ret) {
-			g_critical("Invalid layout, proper choices are QWERTY, QWERTY_REV, QWERTY_UK, QWERTY_UK_REV, QWERTZ, AZERTY and DVORAK.");
+			g_critical("Invalid layout, proper choices are QWERTY.");
 			exit(EX_USAGE);
 		}
 	}
