@@ -35,6 +35,8 @@
 #include <string.h>
 #include <gtk/gtk.h>
 #include <gdk/gdkkeysyms.h>
+#include <cairo.h>
+#include <math.h>
 
 #include "pianokeyboard.h"
 
@@ -74,20 +76,15 @@ draw_keyboard_cue(PianoKeyboard *pk)
 	    h - 9, pk->notes[last_note_in_higher_row].x + w - 3, h - 9);
 }
 
-static void 
+static void
 draw_note(PianoKeyboard *pk, int note)
 {
+	GtkWidget *widget = GTK_WIDGET(pk);
 	if (note < pk->min_note)
 		return;
 	if (note > pk->max_note)
 		return;
-	int is_white, x, w, h;
-
-	GdkColor black = {0, 0, 0, 0};
-	GdkColor white = {0, 65535, 65535, 65535};
-
-	GdkGC *gc = GTK_WIDGET(pk)->style->fg_gc[0];
-	GtkWidget *widget;
+	int is_white, x, w, h, pressed;
 
 	is_white = pk->notes[note].white;
 
@@ -95,28 +92,23 @@ draw_note(PianoKeyboard *pk, int note)
 	w = pk->notes[note].w;
 	h = pk->notes[note].h;
 
-	if (pk->notes[note].pressed || pk->notes[note].sustained)
-		is_white = !is_white;
-
-	if (is_white)
-		gdk_gc_set_rgb_fg_color(gc, &white);
-	else
-		gdk_gc_set_rgb_fg_color(gc, &black);
-
-	gdk_draw_rectangle(GTK_WIDGET(pk)->window, gc, TRUE, x, 0, w, h);
-	gdk_gc_set_rgb_fg_color(gc, &black);
-	gdk_draw_rectangle(GTK_WIDGET(pk)->window, gc, FALSE, x, 0, w, h);
-
-	if (pk->enable_keyboard_cue)
-		draw_keyboard_cue(pk);
-
-	/* We need to redraw black keys that partially obscure the white one. */
+	pressed = (int)(pk->notes[note].pressed || pk->notes[note].sustained);
+	
+	if (is_white) {
+		piano_keyboard_draw_white_key(widget, x, 0, w, h, pressed, pk->notes[note].velocity);
+	} else {
+		piano_keyboard_draw_black_key(widget, x, 0, w, h, pressed, pk->notes[note].velocity);
+	}
+	
 	if (note < NNOTES - 2 && !pk->notes[note + 1].white)
 		draw_note(pk, note + 1);
-
+	
 	if (note > 0 && !pk->notes[note - 1].white)
 		draw_note(pk, note - 1);
-
+			
+	if (pk->enable_keyboard_cue)
+		draw_keyboard_cue(pk);
+		
 	/*
 	 * XXX: This doesn't really belong here.  Originally I wanted to pack PianoKeyboard into GtkFrame
 	 * packed into GtkAlignment.  I failed to make it behave the way I want.  GtkFrame would need
@@ -129,7 +121,7 @@ draw_note(PianoKeyboard *pk, int note)
 	    widget->allocation.height);
 }
 
-static int 
+static int
 press_key(PianoKeyboard *pk, int key)
 {
 	assert(key >= 0);
@@ -147,6 +139,7 @@ press_key(PianoKeyboard *pk, int key)
 		pk->notes[key].sustained = 0;
 
 	pk->notes[key].pressed = 1;
+	pk->notes[key].velocity = pk->current_velocity;
 
 	g_signal_emit_by_name(GTK_WIDGET(pk), "note-on", key);
 	draw_note(pk, key);
@@ -154,7 +147,7 @@ press_key(PianoKeyboard *pk, int key)
 	return (1);
 }
 
-static int 
+static int
 release_key(PianoKeyboard *pk, int key)
 {
 	assert(key >= 0);
@@ -179,7 +172,7 @@ release_key(PianoKeyboard *pk, int key)
 	return (1);
 }
 
-static void 
+static void
 stop_unsustained_notes(PianoKeyboard *pk)
 {
 	int i;
@@ -193,7 +186,7 @@ stop_unsustained_notes(PianoKeyboard *pk)
 	}
 }
 
-static void 
+static void
 stop_sustained_notes(PianoKeyboard *pk)
 {
 	int i;
@@ -209,27 +202,24 @@ stop_sustained_notes(PianoKeyboard *pk)
 }
 
 static int
-key_binding(PianoKeyboard *pk, const char *key)
+key_binding(PianoKeyboard *pk, guint16 key)
 {
-	gpointer notused, note;
-	gboolean found;
-
 	assert(pk->key_bindings != NULL);
 
-	found = g_hash_table_lookup_extended(pk->key_bindings, key, &notused, &note);
+	if (key >= pk->key_bindings->len)
+		return (0);
 
-	if (!found)
-		return (-1);
-
-	return ((long)note);
+	return (g_array_index(pk->key_bindings, int, key));
 }
 
 static void
-bind_key(PianoKeyboard *pk, const char *key, long note)
+bind_key(PianoKeyboard *pk, guint key, int note)
 {
 	assert(pk->key_bindings != NULL);
 
-	g_hash_table_insert(pk->key_bindings, (gpointer)key, (gpointer)note);
+	if (key >= pk->key_bindings->len)
+		g_array_set_size(pk->key_bindings, key + 1);
+	g_array_index(pk->key_bindings, int, key) = note;
 }
 
 static void
@@ -237,258 +227,72 @@ clear_notes(PianoKeyboard *pk)
 {
 	assert(pk->key_bindings != NULL);
 
-	g_hash_table_remove_all(pk->key_bindings);
+	g_array_set_size(pk->key_bindings, 0);
 }
 
-static void 
+static void
 bind_keys_qwerty(PianoKeyboard *pk)
 {
 	clear_notes(pk);
 
-	/* Lower keyboard row - "zxcvbnm". */
-	bind_key(pk, "z", 12);	/* C0 */
-	bind_key(pk, "s", 13);
-	bind_key(pk, "x", 14);
-	bind_key(pk, "d", 15);
-	bind_key(pk, "c", 16);
-	bind_key(pk, "v", 17);
-	bind_key(pk, "g", 18);
-	bind_key(pk, "b", 19);
-	bind_key(pk, "h", 20);
-	bind_key(pk, "n", 21);
-	bind_key(pk, "j", 22);
-	bind_key(pk, "m", 23);
+	/* Lower keyboard row - "zxcvbnm". 52 is z, 38 is a. */
+	bind_key(pk, 52, 12);	/* C0 */
+	bind_key(pk, 39, 13);
+	bind_key(pk, 53, 14);
+	bind_key(pk, 40, 15);
+	bind_key(pk, 54, 16);
+	bind_key(pk, 55, 17);
+	bind_key(pk, 42, 18);
+	bind_key(pk, 56, 19);
+	bind_key(pk, 43, 20);
+	bind_key(pk, 57, 21);
+	bind_key(pk, 44, 22);
+	bind_key(pk, 58, 23);
 
-	/* Upper keyboard row, first octave - "qwertyu". */
-	bind_key(pk, "q", 24); /* C1 */
-	bind_key(pk, "2", 25);
-	bind_key(pk, "w", 26);
-	bind_key(pk, "3", 27);
-	bind_key(pk, "e", 28);
-	bind_key(pk, "r", 29);
-	bind_key(pk, "5", 30);
-	bind_key(pk, "t", 31);
-	bind_key(pk, "6", 32);
-	bind_key(pk, "y", 33);
-	bind_key(pk, "7", 34);
-	bind_key(pk, "u", 35);
+	/* Map the remaining keys. This overlaps with qwe. */
+	bind_key(pk, 59, 24);
+	bind_key(pk, 46, 25);
+	bind_key(pk, 60, 26);
+	bind_key(pk, 47, 27);
+	bind_key(pk, 61, 28);
+
+	/* Upper keyboard row, first octave - "qwertyu". 24 is q, 10 is 1. */
+	bind_key(pk, 24, 24);
+	bind_key(pk, 11, 25);
+	bind_key(pk, 25, 26);
+	bind_key(pk, 12, 27);
+	bind_key(pk, 26, 28);
+	bind_key(pk, 27, 29);
+	bind_key(pk, 14, 30);
+	bind_key(pk, 28, 31);
+	bind_key(pk, 15, 32);
+	bind_key(pk, 29, 33);
+	bind_key(pk, 16, 34);
+	bind_key(pk, 30, 35);
 
 	/* Upper keyboard row, the rest - "iop". */
-	bind_key(pk, "i", 36); /* C2 */
-	bind_key(pk, "9", 37);
-	bind_key(pk, "o", 38);
-	bind_key(pk, "0", 39);
-	bind_key(pk, "p", 40);
+	bind_key(pk, 31, 36);
+	bind_key(pk, 18, 37);
+	bind_key(pk, 32, 38);
+	bind_key(pk, 19, 39);
+	bind_key(pk, 33, 40);
+
+	/* We might as well bind these too: "[=]\" */
+	bind_key(pk, 34, 41);
+	bind_key(pk, 21, 42);
+	bind_key(pk, 35, 43);
+	bind_key(pk, 51, 45); /* yes, really, at least here... */
 }
 
-static void 
-bind_keys_qwerty_uk(PianoKeyboard *pk)
-{
-	bind_keys_qwerty(pk);
-
-	/* Lower keyboard row - "zxcvbnm". */
-	bind_key(pk, "backslash", 11); /* B0 */
-	/* ... */
-	bind_key(pk, "comma", 24); /* C1 */
-	bind_key(pk, "l", 25);
-	bind_key(pk, "period", 26);
-	bind_key(pk, "semicolon", 27);
-	bind_key(pk, "slash", 28);
-
-	/* Upper keyboard row, the rest - "iop". */
-	bind_key(pk, "bracketleft", 41); /* F6 */
-	bind_key(pk, "equal", 42);
-	bind_key(pk, "bracketright", 43);
-}
-static void 
-bind_keys_qwerty_rev(PianoKeyboard *pk)
-{
-	clear_notes(pk);
-
-	/* Lower keyboard row - "zxcvbnm". */
-	bind_key(pk, "z", 24);	/* C1 */
-	bind_key(pk, "s", 25);
-	bind_key(pk, "x", 26);
-	bind_key(pk, "d", 27);
-	bind_key(pk, "c", 28);
-	bind_key(pk, "v", 29);
-	bind_key(pk, "g", 30);
-	bind_key(pk, "b", 31);
-	bind_key(pk, "h", 32);
-	bind_key(pk, "n", 33);
-	bind_key(pk, "j", 34);
-	bind_key(pk, "m", 35);
-
-	/* Upper keyboard row, first octave - "qwertyu". */
-	bind_key(pk, "q", 12); /* C0 */
-	bind_key(pk, "2", 13);
-	bind_key(pk, "w", 14);
-	bind_key(pk, "3", 15);
-	bind_key(pk, "e", 16);
-	bind_key(pk, "r", 17);
-	bind_key(pk, "5", 18);
-	bind_key(pk, "t", 19);
-	bind_key(pk, "6", 20);
-	bind_key(pk, "y", 21);
-	bind_key(pk, "7", 22);
-	bind_key(pk, "u", 23);
-
-	/* Upper keyboard row, the rest - "iop". */
-	bind_key(pk, "i", 24); /* C1 */
-	bind_key(pk, "9", 25);
-	bind_key(pk, "o", 26);
-	bind_key(pk, "0", 27);
-	bind_key(pk, "p", 28);
-}
-
-static void 
-bind_keys_qwerty_uk_rev(PianoKeyboard *pk)
-{
-	bind_keys_qwerty_rev(pk);
-
-	/* Lower keyboard row - "zxcvbnm". */
-	bind_key(pk, "backslash", 23); /* B-1 */
-	/* ... */
-	bind_key(pk, "comma", 36); /* C2 */
-	bind_key(pk, "l", 37);
-	bind_key(pk, "period", 38);
-	bind_key(pk, "semicolon", 39);
-	bind_key(pk, "slash", 40);
-
-	/* Upper keyboard row, the rest - "iop". */
-	bind_key(pk, "bracketleft", 29);
-	bind_key(pk, "equal", 30);
-	bind_key(pk, "bracketright", 31);
-}
-
-static void 
-bind_keys_qwertz(PianoKeyboard *pk)
-{
-	bind_keys_qwerty(pk);
-
-	/* The only difference between QWERTY and QWERTZ is that the "y" and "z" are swapped together. */
-	bind_key(pk, "y", 12);
-	bind_key(pk, "z", 33);
-}
-
-static void
-bind_keys_azerty(PianoKeyboard *pk)
-{
-	clear_notes(pk);
-
-	/* Lower keyboard row - "wxcvbn,". */
-	bind_key(pk, "w", 12);	/* C0 */
-	bind_key(pk, "s", 13);
-	bind_key(pk, "x", 14);
-	bind_key(pk, "d", 15);
-	bind_key(pk, "c", 16);
-	bind_key(pk, "v", 17);
-	bind_key(pk, "g", 18);
-	bind_key(pk, "b", 19);
-	bind_key(pk, "h", 20);
-	bind_key(pk, "n", 21);
-	bind_key(pk, "j", 22);
-	bind_key(pk, "comma", 23);
-
-	/* Upper keyboard row, first octave - "azertyu". */
-	bind_key(pk, "a", 24);
-	bind_key(pk, "eacute", 25);
-	bind_key(pk, "z", 26);
-	bind_key(pk, "quotedbl", 27);
-	bind_key(pk, "e", 28);
-	bind_key(pk, "r", 29);
-	bind_key(pk, "parenleft", 30);
-	bind_key(pk, "t", 31);
-	bind_key(pk, "minus", 32);
-	bind_key(pk, "y", 33);
-	bind_key(pk, "egrave", 34);
-	bind_key(pk, "u", 35);
-
-	/* Upper keyboard row, the rest - "iop". */
-	bind_key(pk, "i", 36);
-	bind_key(pk, "ccedilla", 37);
-	bind_key(pk, "o", 38);
-	bind_key(pk, "agrave", 39);
-	bind_key(pk, "p", 40);
-}
-
-static void 
-bind_keys_dvorak(PianoKeyboard *pk)
-{
-	clear_notes(pk);
-
-	/* Lower keyboard row - ";qjkxbm". */
-	bind_key(pk, "semicolon", 12); /* C0 */
-	bind_key(pk, "o", 13);
-	bind_key(pk, "q", 14);
-	bind_key(pk, "e", 15);
-	bind_key(pk, "j", 16);
-	bind_key(pk, "k", 17);
-	bind_key(pk, "i", 18);
-	bind_key(pk, "x", 19);
-	bind_key(pk, "d", 20);
-	bind_key(pk, "b", 21);
-	bind_key(pk, "h", 22);
-	bind_key(pk, "m", 23);
-	bind_key(pk, "w", 24); /* overlaps with upper row */
-	bind_key(pk, "n", 25);
-	bind_key(pk, "v", 26);
-	bind_key(pk, "s", 27);
-	bind_key(pk, "z", 28);
-
-	/* Upper keyboard row, first octave - "',.pyfg". */
-	bind_key(pk, "apostrophe", 24);
-	bind_key(pk, "2", 25);
-	bind_key(pk, "comma", 26);
-	bind_key(pk, "3", 27);
-	bind_key(pk, "period", 28);
-	bind_key(pk, "p", 29);
-	bind_key(pk, "5", 30);
-	bind_key(pk, "y", 31);
-	bind_key(pk, "6", 32);
-	bind_key(pk, "f", 33);
-	bind_key(pk, "7", 34);
-	bind_key(pk, "g", 35);
-
-	/* Upper keyboard row, the rest - "crl". */
-	bind_key(pk, "c", 36);
-	bind_key(pk, "9", 37);
-	bind_key(pk, "r", 38);
-	bind_key(pk, "0", 39);
-	bind_key(pk, "l", 40);
-	bind_key(pk, "slash", 41); /* extra F */
-	bind_key(pk, "bracketright", 42);
-	bind_key(pk, "equal", 43);
-
-}
-
-static gint 
+static gint
 keyboard_event_handler(GtkWidget *mk, GdkEventKey *event, gpointer notused)
 {
 	int note;
-	char *key;
-	guint keyval;
-	GdkKeymapKey kk;
 	PianoKeyboard *pk = PIANO_KEYBOARD(mk);
 
-	/* We're not using event->keyval, because we need keyval with level set to 0.
-	   E.g. if user holds Shift and presses '7', we want to get a '7', not '&'. */
-	kk.keycode = event->hardware_keycode;
-	kk.level = 0;
-	kk.group = 0;
+	note = key_binding(pk, event->hardware_keycode);
 
-	keyval = gdk_keymap_lookup_key(NULL, &kk);
-
-	key = gdk_keyval_name(gdk_keyval_to_lower(keyval));
-
-	if (key == NULL) {
-		g_message("gtk_keyval_name() returned NULL; please report this.");
-		return (FALSE);
-	}
-
-	note = key_binding(pk, key);
-
-	if (note < 0) {
+	if (note <= 0) {
 		/* Key was not bound.  Maybe it's one of the keys handled in jack-keyboard.c. */
 		return (FALSE);
 	}
@@ -508,7 +312,7 @@ keyboard_event_handler(GtkWidget *mk, GdkEventKey *event, gpointer notused)
 	return (TRUE);
 }
 
-static int 
+static int
 get_note_for_xy(PianoKeyboard *pk, int x, int y)
 {
 	int height, note;
@@ -536,7 +340,7 @@ get_note_for_xy(PianoKeyboard *pk, int x, int y)
 	return (-1);
 }
 
-static gboolean 
+static gboolean
 mouse_button_event_handler(PianoKeyboard *pk, GdkEventButton *event, gpointer notused)
 {
 	int x, y, note;
@@ -578,7 +382,7 @@ mouse_button_event_handler(PianoKeyboard *pk, GdkEventButton *event, gpointer no
 	return (TRUE);
 }
 
-static gboolean 
+static gboolean
 mouse_motion_event_handler(PianoKeyboard *pk, GdkEventMotion *event, gpointer notused)
 {
 	int note;
@@ -589,7 +393,7 @@ mouse_motion_event_handler(PianoKeyboard *pk, GdkEventMotion *event, gpointer no
 	note = get_note_for_xy(pk, event->x, event->y);
 
 	if (note != pk->note_being_pressed_using_mouse && note >= 0) {
-		
+
 		if (pk->note_being_pressed_using_mouse >= 0)
 			release_key(pk, pk->note_being_pressed_using_mouse);
 		press_key(pk, note);
@@ -611,7 +415,7 @@ piano_keyboard_expose(GtkWidget *widget, GdkEventExpose *event)
 	return (TRUE);
 }
 
-static void 
+static void
 piano_keyboard_size_request(GtkWidget *widget, GtkRequisition *requisition)
 {
 	requisition->width = PIANO_KEYBOARD_DEFAULT_WIDTH;
@@ -676,7 +480,7 @@ recompute_dimensions(PianoKeyboard *pk)
 			    (white_key * key_width) -
 			    (black_key_width * black_key_left_shift(note));
 			pk->notes[note].w = black_key_width;
-			pk->notes[note].h = (height * 2) / 3;
+			pk->notes[note].h = (height * 3) / 5;
 			pk->notes[note].white = 0;
 			continue;
 		}
@@ -720,7 +524,7 @@ piano_keyboard_class_init(PianoKeyboardClass *klass)
 	    G_TYPE_FROM_CLASS (klass), G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
 	    0, NULL, NULL, g_cclosure_marshal_VOID__INT, G_TYPE_NONE, 1, G_TYPE_INT);
 
-	widget_klass = (GtkWidgetClass*)klass;	
+	widget_klass = (GtkWidgetClass*)klass;
 
 	widget_klass->expose_event = piano_keyboard_expose;
 	widget_klass->size_request = piano_keyboard_size_request;
@@ -778,7 +582,8 @@ piano_keyboard_new(void)
 	pk->octave = 4;
 	pk->note_being_pressed_using_mouse = -1;
 	memset((void *)pk->notes, 0, sizeof(struct Note) * NNOTES);
-	pk->key_bindings = g_hash_table_new(g_str_hash, g_str_equal);
+	/* 255 here is a random value larger than the highest key we bind. */
+	pk->key_bindings = g_array_sized_new(FALSE, TRUE, sizeof(int), 255);
 	pk->min_note = PIANO_MIN_NOTE;
 	pk->max_note = PIANO_MAX_NOTE;
 	bind_keys_qwerty(pk);
@@ -801,7 +606,7 @@ piano_keyboard_sustain_press(PianoKeyboard *pk)
 	}
 }
 
-void	
+void
 piano_keyboard_sustain_release(PianoKeyboard *pk)
 {
 	if (pk->maybe_stop_sustained_notes)
@@ -811,10 +616,11 @@ piano_keyboard_sustain_release(PianoKeyboard *pk)
 }
 
 void
-piano_keyboard_set_note_on(PianoKeyboard *pk, int note)
+piano_keyboard_set_note_on(PianoKeyboard *pk, int note, int vel)
 {
 	if (pk->notes[note].pressed == 0) {
 		pk->notes[note].pressed = 1;
+		pk->notes[note].velocity = vel;
 		draw_note(pk, note);
 	}
 }
@@ -845,24 +651,6 @@ piano_keyboard_set_keyboard_layout(PianoKeyboard *pk, const char *layout)
 	if (!strcasecmp(layout, "QWERTY")) {
 		bind_keys_qwerty(pk);
 
-	} else if (!strcasecmp(layout, "QWERTY_REV")) {
-		bind_keys_qwerty_rev(pk);
-
-	} else if (!strcasecmp(layout, "QWERTY_UK")) {
-		bind_keys_qwerty_uk(pk);
-
-	} else if (!strcasecmp(layout, "QWERTY_UK_REV")) {
-		bind_keys_qwerty_uk_rev(pk);
-
-	} else if (!strcasecmp(layout, "QWERTZ")) {
-		bind_keys_qwertz(pk);
-
-	} else if (!strcasecmp(layout, "AZERTY")) {
-		bind_keys_azerty(pk);
-
-	} else if (!strcasecmp(layout, "DVORAK")) {
-		bind_keys_dvorak(pk);
-
 	} else {
 		/* Unknown layout name. */
 		return (TRUE);
@@ -879,3 +667,137 @@ piano_keyboard_enable_all_midi_notes(PianoKeyboard* pk)
 	recompute_dimensions(pk);
 }
 
+void
+piano_keyboard_draw_white_key (GtkWidget *widget, int x, int y, int w, int h, int pressed, int vel)
+{
+	cairo_pattern_t *pat;
+	GdkWindow *window = widget->window;
+	cairo_t *c = gdk_cairo_create(GDK_DRAWABLE(window));
+	cairo_set_line_join(c, CAIRO_LINE_JOIN_MITER);
+	cairo_set_line_width(c, 1);
+	
+	cairo_rectangle(c, x, y, w, h);
+	cairo_clip_preserve(c);
+	
+	pat = cairo_pattern_create_linear (x, y, x, y + h);
+	cairo_pattern_add_color_stop_rgb (pat, 0.0, 0.25, 0.25, 0.2);
+	cairo_pattern_add_color_stop_rgb (pat, 0.1, 0.957, 0.914, 0.925);
+	cairo_pattern_add_color_stop_rgb (pat, 1.0, 0.796, 0.787, 0.662);
+	cairo_set_source(c, pat);
+	cairo_fill(c);
+	
+	cairo_move_to(c, x + 0.5, y);
+	cairo_line_to(c, x + 0.5, y + h);
+	cairo_set_source_rgba(c, 1, 1, 1, 0.75);
+	cairo_stroke(c);
+	
+	cairo_move_to(c, x + w - 0.5, y);
+	cairo_line_to(c, x + w - 0.5, y + h);
+	cairo_set_source_rgba(c, 0, 0, 0, 0.5);
+	cairo_stroke(c);
+	
+	if (pressed)
+		piano_keyboard_draw_pressed(c, x, y, w, h, vel);
+		
+	piano_keyboard_draw_key_shadow(c, x, y, w, h);
+	
+	cairo_destroy(c);
+}
+
+void
+piano_keyboard_draw_black_key (GtkWidget *widget, int x, int y, int w, int h, int pressed, int vel)
+{
+	cairo_pattern_t *pat;
+	GdkWindow *window = widget->window;
+	cairo_t *c = gdk_cairo_create(GDK_DRAWABLE(window));
+	cairo_set_line_join(c, CAIRO_LINE_JOIN_MITER);
+	cairo_set_line_width(c, 1);
+	
+	cairo_rectangle(c, x, y, w, h);
+	cairo_clip_preserve(c);
+	
+	pat = cairo_pattern_create_linear (x, y, x, y + h);
+	cairo_pattern_add_color_stop_rgb (pat, 0.0, 0, 0, 0);
+	cairo_pattern_add_color_stop_rgb (pat, 0.1, 0.27, 0.27, 0.27);
+	cairo_pattern_add_color_stop_rgb (pat, 1.0, 0, 0, 0);
+	cairo_set_source(c, pat);
+	cairo_fill(c);
+	
+	pat = cairo_pattern_create_linear (x + 1, y, x + 1, y + h - w);
+	cairo_pattern_add_color_stop_rgb (pat, 0.0, 0, 0, 0);
+	cairo_pattern_add_color_stop_rgb (pat, 0.1, 0.55, 0.55, 0.55);
+	cairo_pattern_add_color_stop_rgb (pat, 0.5, 0.45, 0.45, 0.45);
+	cairo_pattern_add_color_stop_rgb (pat, 0.5001, 0.35, 0.35, 0.35);
+	cairo_pattern_add_color_stop_rgb (pat, 1.0, 0.25, 0.25, 0.25);
+	cairo_set_source(c, pat);
+	cairo_rectangle(c, x + 1, y, w - 2, y + h - w);
+	cairo_fill(c);
+	
+	if (pressed)
+		piano_keyboard_draw_pressed(c, x, y, w, h, vel);
+	
+	piano_keyboard_draw_key_shadow(c, x, y, w, h);
+	
+	cairo_destroy(c);
+}
+
+void
+piano_keyboard_draw_pressed (cairo_t *c, int x, int y, int w, int h, int vel)
+{
+	float m = w * .15; // margin
+	float s = w - m * 2.; // size
+	float _vel = ((float)vel / 127.);
+	float hue = _vel * 140 + 220; // hue 220 .. 360 - blue over pink to red
+	float sat = .5 + _vel * 0.3; // saturation 0.5 .. 0.8
+	float val = 1. - _vel * 0.2; // lightness 1.0 .. 0.8
+	cairo_rectangle(c, x + m, y + h - m - s * 2, s, s * 2);
+	hsv HSV = {hue, sat, val};
+	rgb RGB = hsv2rgb(HSV);
+	cairo_set_source_rgb(c, RGB.r, RGB.g, RGB.b);
+	cairo_fill(c);
+}
+
+void
+piano_keyboard_draw_key_shadow (cairo_t *c, int x, int y, int w, int h)
+{
+	cairo_pattern_t *pat;
+	pat = cairo_pattern_create_linear (x, y, x, y + (int)(h * 0.2));
+	cairo_pattern_add_color_stop_rgba (pat, 0.0, 0, 0, 0, 0.4);
+	cairo_pattern_add_color_stop_rgba (pat, 1.0, 0, 0, 0, 0);
+	cairo_rectangle(c, x, y, w, (int)(h * 0.2));
+	cairo_set_source(c, pat);
+	cairo_fill(c);
+}
+
+rgb
+hsv2rgb(hsv HSV)
+{
+	rgb RGB;
+	double H = HSV.h, S = HSV.s, V = HSV.v,
+			P, Q, T,
+			fract;
+
+	(H == 360.)?(H = 0.):(H /= 60.);
+	fract = H - floor(H);
+
+	P = V*(1. - S);
+	Q = V*(1. - S*fract);
+	T = V*(1. - S*(1. - fract));
+
+	if (0. <= H && H < 1.)
+		RGB = (rgb){.r = V, .g = T, .b = P};
+	else if (1. <= H && H < 2.)
+		RGB = (rgb){.r = Q, .g = V, .b = P};
+	else if (2. <= H && H < 3.)
+		RGB = (rgb){.r = P, .g = V, .b = T};
+	else if (3. <= H && H < 4.)
+		RGB = (rgb){.r = P, .g = Q, .b = V};
+	else if (4. <= H && H < 5.)
+		RGB = (rgb){.r = T, .g = P, .b = V};
+	else if (5. <= H && H < 6.)
+		RGB = (rgb){.r = V, .g = P, .b = Q};
+	else
+		RGB = (rgb){.r = 0., .g = 0., .b = 0.};
+
+	return RGB;
+}
